package ru.ulstu.fc.rule.service;

import com.fuzzylite.Engine;
import com.fuzzylite.activation.General;
import com.fuzzylite.defuzzifier.WeightedAverage;
import com.fuzzylite.norm.s.Maximum;
import com.fuzzylite.norm.t.AlgebraicProduct;
import com.fuzzylite.norm.t.Minimum;
import com.fuzzylite.rule.Rule;
import com.fuzzylite.rule.RuleBlock;
import com.fuzzylite.term.Triangle;
import com.fuzzylite.variable.InputVariable;
import com.fuzzylite.variable.OutputVariable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;
import ru.ulstu.fc.rule.model.FuzzyTerm;
import ru.ulstu.fc.rule.model.OutputValue;
import ru.ulstu.fc.rule.model.Variable;

import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Service
public class FuzzyInferenceService {
    private final static Logger LOG = LoggerFactory.getLogger(FuzzyInferenceService.class);
    private final static String OUTPUT_VARIABLE_NAME = "кредит";
    private final static String RULE_TEMPLATE = "if %s is %s and %s is %s then "
            + OUTPUT_VARIABLE_NAME
            + " is %s";
    private final static String NO_RESULT = "Нет результата";
    private final Engine fuzzyEngine;

    public FuzzyInferenceService(Engine fuzzyEngine) {
        this.fuzzyEngine = fuzzyEngine;
    }

    private List<String> getDemoRules() {
        return List.of(
                String.format(RULE_TEMPLATE, "возраст", "молодой", "доход", "высокий", "средний"),
                String.format(RULE_TEMPLATE, "возраст", "средний", "доход", "высокий", "большой"),
                String.format(RULE_TEMPLATE, "возраст", "старый", "доход", "средний", "средний")
        );
    }

    private InputVariable getInputVariable(Variable variable) {
        final InputVariable input = new InputVariable();
        input.setName(variable.getName());
        input.setDescription("");
        input.setRange(0, variable.getFuzzyTerms().get(variable.getFuzzyTerms().size() - 1).getCrispValue());
        input.setEnabled(true);
        input.setLockValueInRange(false);
        double prev = 0;
        for (int i = 0; i < variable.getFuzzyTerms().size(); i++) {
            Triangle term = new Triangle(variable.getFuzzyTerms().get(i).getDescription(),
                    prev,
                    variable.getFuzzyTerms().get(i).getCrispValue(),
                    variable.getFuzzyTerms().get(i).getCrispValue() + variable.getFuzzyTerms().get(i).getCrispValue() - prev);
            prev = term.getVertexB();
            input.addTerm(term);
        }
        return input;
    }

    private <T extends Enum<T>> OutputVariable getOutputVariable(Variable variable) {
        final OutputVariable output = new OutputVariable();
        output.setName(variable.getName());
        output.setDescription("");
        output.setRange(0, variable.getFuzzyTerms().get(variable.getFuzzyTerms().size() - 1).getCrispValue());
        output.setEnabled(true);
        output.setAggregation(new Maximum());
        output.setDefuzzifier(new WeightedAverage());
        output.setDefaultValue(Double.NaN);
        output.setLockValueInRange(false);
        double prev = 0;
        for (int i = 0; i < variable.getFuzzyTerms().size(); i++) {
            Triangle term = new Triangle(
                    variable.getFuzzyTerms().get(i).getDescription(),
                    prev,
                    variable.getFuzzyTerms().get(i).getCrispValue(),
                    variable.getFuzzyTerms().get(i).getCrispValue() + variable.getFuzzyTerms().get(i).getCrispValue() - prev);
            prev = term.getVertexB();
            output.addTerm(term);
        }
        return output;
    }

    private RuleBlock getRuleBlock(Engine engine,
                                   List<String> rules,
                                   List<Variable> inputVariables,
                                   Variable outputVariable) {
        inputVariables.stream().map(this::getInputVariable).forEach(engine::addInputVariable);
        engine.addOutputVariable(getOutputVariable(outputVariable));

        RuleBlock mamdani = new RuleBlock();
        mamdani.setName("mamdani");
        mamdani.setDescription("");
        mamdani.setEnabled(true);
        mamdani.setConjunction(new Minimum());
        mamdani.setImplication(new AlgebraicProduct());
        mamdani.setActivation(new General());
        rules.forEach(r -> mamdani.addRule(Rule.parse(r, engine)));
        mamdani.addRule(new Rule());
        return mamdani;
    }

    private List<OutputValue> getConsequent(Engine engine, Map<String, Double> variableValues) {
        OutputVariable outputVariable = engine.getOutputVariable(OUTPUT_VARIABLE_NAME);
        for (Map.Entry<String, Double> variableValue : variableValues.entrySet()) {
            InputVariable inputVariable = engine.getInputVariable(variableValue.getKey());
            inputVariable.setValue(variableValue.getValue());
        }
        engine.process();
        if (outputVariable != null) {
            LOG.info("Output: {}", outputVariable.getValue());
        }
        return Double.isNaN(outputVariable.getValue())
                ? List.of(new OutputValue(NO_RESULT, 0.0))
                : outputVariable.fuzzyOutput()
                .getTerms()
                .stream()
                .map(t -> new OutputValue(t.getTerm().getName(), t.getDegree()))
                .collect(Collectors.toList());
    }

    public List<OutputValue> getFuzzyInference(Map<String, Double> vals) {
        return getFuzzyInference(getDemoRules(), vals,
                List.of(new Variable("возраст", List.of(
                                new FuzzyTerm("молодой", 35.0),
                                new FuzzyTerm("средний", 60.0),
                                new FuzzyTerm("старый", 100.0))
                        ),
                        new Variable("доход", List.of(
                                new FuzzyTerm("небольшой", 35000.0),
                                new FuzzyTerm("средний", 100000.0),
                                new FuzzyTerm("высокий", 500000.0))
                        )
                ),
                new Variable("кредит", List.of(
                        new FuzzyTerm("небольшой", 20000.0),
                        new FuzzyTerm("средний", 100000.0),
                        new FuzzyTerm("большой", 1000000.0)))
        );
    }

    public List<OutputValue> getFuzzyInference(List<String> rules,
                                               Map<String, Double> values,
                                               List<Variable> inputVariables,
                                               Variable outputVariable) {
        fuzzyEngine.getRuleBlocks().clear();
        fuzzyEngine.addRuleBlock(getRuleBlock(fuzzyEngine, rules, inputVariables, outputVariable));
        return getConsequent(fuzzyEngine, values);
    }
}
